"""
========================================
  Introduction to Computer Programming
========================================

----------------------------------------------------
  examples in preparation for lab 4:
    * using string subscripting
    * functions that call other functions
    * demystifying common string operations
----------------------------------------------------
"""


def has_any_digits(s):
    """Returns True if any of the characters in string s are digits.

    >>> has_any_digits('')
    False
    >>> has_any_digits('9')
    True
    >>> has_any_digits('65537')
    True
    >>> has_any_digits('There is 1 digit in here!')
    True
    >>> has_any_digits('Nothing to see here.')
    False
    """
    b = False
    i = 0
    while not b and i < len(s):
        b = '0' <= s[i] and s[i] <= '9'
        i += 1
    return b


# very common string operation: check if a smaller string (in this case a
# single character) is in a string at all; we considered a version of this
# earlier using a for loop, but not only did that have the inefficiency of the
# definite loop, it required us to "remember" that we had seen the character.
# This while loop with subscripting version is definitely preferable.
def has(s, x):
    """
    Returns True if character x occurs in string s, False otherwise. Assumes x
    is a character (string of length 1).

    >>> has('Hello world!', 'w')
    True
    >>> has('', 'T')
    False
    >>> has('Hello world!', 'W')   # case sensitive
    False
    """
    found = False
    i = 0
    while not found and i < len(s):
        found = s[i] == x
        i += 1
    return found


# familiar from previous lab
def is_upper_char(ch):
    """Returns True if ch is a single uppercase letter, False otherwise.

    >>> is_upper_char('P')
    True
    >>> is_upper_char('p')
    False
    >>> is_upper_char('!')
    False
    >>> is_upper_char('CAPS')
    False
    """
    b = len(ch) == 1 and 'A' <= ch and ch <= 'Z'
    return b


def is_all_upper(s):
    """
    Returns True if s is a nonempty string consisting entirely of uppercase
    letters.

    >>> is_all_upper('')
    False
    >>> is_all_upper('NOT!')   # ! is not an uppercase letter
    False
    >>> is_all_upper('MikeSiff')
    False
    >>> is_all_upper('MIKE SIFF')   # space is not an uppercase letter
    False
    >>> is_all_upper('MIKESIFF')
    True
    """
    b = len(s) > 0
    i = 0
    while b and i < len(s):
        b = is_upper_char(s[i])
        i += 1
    return b


def string_equals(s, t):
    """
    Returns True if the two strings are equal, False otherwise. Equivalent in
    effect to Python's built-in == operator on strings.

    >>> string_equals('Apple', 'Apple')
    True
    >>> string_equals('Apples', 'Apple')
    False
    >>> string_equals('Apple', 'Apples')
    False
    >>> string_equals('Apple', 'apple')   # case sensitive!
    False
    """
    eq_so_far = len(s) == len(t)
    i = 0
    while eq_so_far and i < len(s):
        eq_so_far = s[i] == t[i]
        i += 1
    return eq_so_far


# familiar idea from previous lab
# observe it is polymorphic: it works on any x, y for which < is defined
def compare(x, y):
    """
    Returns -1, 0, or 1, respectively depending if x is <, ==, or < than y.

    >>> compare(3, 4)
    -1
    >>> compare('hello', 'Hello')   # case sensitive!
    1
    >>> compare('A' + 'b', 'Ab')
    0
    """
    if x < y:
        z = -1
    elif y < x:
        z = 1
    else:
        z = 0
    return z


def string_order(s, t):
    """
    Returns order (-1, 0, or 1) of string s compared to string t,
    which major assumption that s and t are of the same length.

    >>> string_order('aBc', 'abc')
    -1
    >>> string_order('abc', 'aBc')
    1
    >>> string_order('ABC', 'ABC')
    0
    """
    order = 0
    i = 0
    while order == 0 and i < len(s):
        order = compare(s[i], t[i])
        i += 1
    return order


# we saw a version of this before that awkwardly used for loops
def find(s, x):
    """
    Returns the position (counting from 0!) that character x first occurs in
    string s. Returns -1 if x does not occur in s.

    >>> find('Computer Science!', 'm')   # count from 0
    2
    >>> find('Computer Science!', 'e')   # first occurrence
    6
    >>> find('Computer Science!', 'c')   # case sensitive
    10
    >>> find('Computer Science!', '?')
    -1
    """
    pos = -1
    i = 0
    while pos == -1 and i < len(s):
        if s[i] == x:
            pos = i
        i += 1
    return pos


def substring(s, start, n):
    """
    Returns n characters from s starting at start. Assumes length of s is at
    least start + n.

    >>> substring('inching', 2, 4)
    'chin'
    """
    t = ''
    for i in range(n):
        t += s[start + i]
    return t


# slice is like substring but it uses two positions rather than a length
def string_slice(s, start, stop):
    """
    Returns characters in s from start (inclusive) to stop (exclusive). Assumes
    start and stop are >= 0 and that stop <= len(s).

    >>> string_slice('blinks', 1, 5)
    'link'
    """
    t = ''
    for k in range(start, stop):
        t += s[k]
    return t


# based on underlying ASCII/Unicode value - i.e., ord - but no need to use ord
# here
def find_smallest(s):
    """
    Returns character (string of length 1) in string s that is the "smallest"
    in terms of lexicographic order. Assumes s is not empty.

    >>> find_smallest('what')
    'a'
    >>> find_smallest('iMac')   # uppercase letters are "smaller"
    'M'
    """
    m = s[0]
    for i in range(1, len(s)):
        if s[i] < m:
            m = s[i]
    return m


def string_zip(s, t):
    """
    Returns new string that is alternates characters from s and t. The returned
    string should be twice as long as the shorter of s and t.

    >>> string_zip('ABC','def')
    'AdBeCf'
    >>> string_zip('AB', 'def')
    'AdBe'
    >>> string_zip('ABC', 'de')
    'AdBe'
    """
    z = ''
    for i in range(min(len(s), len(t))):
        z = z + s[i] + t[i]
    return z


# an example of how useful "slicing" can be
def insert(s, x, i):
    """
    Returns new string that is s with character x inserted at position i.
    Assumes 0 <= i <= len(s). If i is length of s then x is simply appended to
    s.

    >>> insert('hello', '!', 5)
    'hello!'
    >>> insert('coolbeans', ' ', 4)
    'cool beans'
    """
    t = string_slice(s, 0, i) + x + string_slice(s, i, len(s))
    return t
